@@ -3,11 +3,13 @@ from functools import partial, wraps
 
 import torch
 import torch.nn.functional as F
-from torch import nn, Tensor, tensor, is_tensor, cat, stft, istft, hann_window, view_as_real, view_as_complex
-from torch.nn import LSTM, Module, ModuleList
+from torch import nn, compiler, Tensor, tensor, is_tensor, cat, stft, istft, hann_window, view_as_real, view_as_complex
+from torch.nn import LSTM, GRU, Module, ModuleList
 
-from einx import add, multiply
-from einops import rearrange, pack, unpack
+from numpy import ndarray
+
+from einx import add, multiply, divide
+from einops import rearrange, repeat, pack, unpack
 from einops.layers.torch import Rearrange
 
 # ein tensor notation:
@@ -22,6 +24,17 @@ from einops.layers.torch import Rearrange
 # constants
 
 LSTM = partial(LSTM, batch_first = True)
+GRU = partial(GRU, batch_first = True)
+
+# disable complex related functions from torch compile
+
+(
+    view_as_real,
+    view_as_complex
+) = tuple(compiler.disable()(fn) for fn in (
+    view_as_real,
+    view_as_complex
+))
 
 # functions
 
@@ -31,6 +44,12 @@ def exists(v):
 def default(v, d):
     return v if exists(v) else d
 
+def divisible_by(num, den):
+    return (num % den) == 0
+
+def identity(t):
+    return t
+
 # residual
 
 def residual(fn):
@@ -42,6 +61,86 @@ def residual(fn):
 
     return decorated
 
+# fft related
+
+class STFT(Module):
+    """
+    need this custom module to address an issue with certain window and no centering in istft in pytorch - https://github.com/pytorch/pytorch/issues/91309
+    this solution was retailored from the working solution used at vocos https://github.com/gemelo-ai/vocos/blob/03c4fcbb321e4b04dd9b5091486eedabf1dc9de0/vocos/spectral_ops.py#L7
+    """
+
+    def __init__(
+        self,
+        n_fft,
+        hop_length,
+        win_length,
+        eps = 1e-11
+    ):
+        super().__init__()
+        self.n_fft = n_fft
+        self.hop_length = hop_length
+        self.win_length = win_length
+
+        window = torch.hann_window(win_length)
+        self.register_buffer('window', window)
+
+        self.eps = eps
+
+    @compiler.disable()
+    def inverse(self, spec):
+        n_fft, hop_length, win_length, window = self.n_fft, self.hop_length, self.win_length, self.window
+
+        batch, freqs, frames = spec.shape
+
+        # inverse FFT
+
+        ifft = torch.fft.irfft(spec, n_fft, dim = 1, norm = 'backward')
+
+        ifft = multiply('b w f, w', ifft, window)
+
+        # overlap and add
+
+        output_size = (frames - 1) * hop_length + win_length
+
+        y = F.fold(
+            ifft,
+            output_size = (1, output_size),
+            kernel_size = (1, win_length),
+            stride = (1, hop_length),
+        )[:, 0, 0]
+
+        # window envelope
+
+        window_sq = repeat(window.square(), 'w -> 1 w t', t = frames)
+
+        window_envelope = F.fold(
+            window_sq,
+            output_size = (1, output_size),
+            kernel_size = (1, win_length),
+            stride = (1, hop_length)
+        )
+
+        window_envelope = rearrange(window_envelope, '1 1 1 n -> n')
+
+        # normalize out
+
+        return divide('b n, n', y, window_envelope.clamp(min = self.eps))
+
+    @compiler.disable()
+    def forward(self, audio):
+
+        stft = torch.stft(
+            audio,
+            n_fft = self.n_fft,
+            win_length = self.win_length,
+            hop_length = self.hop_length,
+            center = False,
+            window = self.window,
+            return_complex = True
+        )
+
+        return stft
+
 # classes
 
 class HSTasNet(Module):
@@ -55,26 +154,35 @@ class HSTasNet(Module):
         overlap_len = 512,
         n_fft = 1024,
         num_sources = 4,    # drums, bass, vocals, other
+        torch_compile = False,
+        use_gru = False
     ):
         super().__init__()
         audio_channels = 2 if stereo else 1
 
         self.audio_channels = audio_channels
-        self.segment_len = segment_len
         self.num_sources = num_sources
 
+        assert overlap_len < segment_len
+
+        self.segment_len = segment_len
+        self.overlap_len = overlap_len
+
+        assert divisible_by(segment_len, 2)
+        self.causal_pad = segment_len // 2
+
         # spec branch encoder stft hparams
 
-        self.stft_kwargs = dict(
+        self.stft = STFT(
             n_fft = n_fft,
             win_length = segment_len,
-            hop_length = overlap_len,
+            hop_length = overlap_len
         )
 
         spec_dim_input = (n_fft // 2 + 1) * 2 * audio_channels
 
         self.spec_encode = nn.Sequential(
-            Rearrange('(b s) f n c -> b n (s f c)', s = audio_channels),
+            Rearrange('(b s) f n ... -> b n (s f ...)', s = audio_channels),
             nn.Linear(spec_dim_input, dim)
         )
 
@@ -87,7 +195,7 @@ class HSTasNet(Module):
 
         self.stereo = stereo
 
-        self.conv_encode = nn.Conv1d(audio_channels, num_basis * 2, segment_len, stride = overlap_len, padding = overlap_len)
+        self.conv_encode = nn.Conv1d(audio_channels, num_basis * 2, segment_len, stride = overlap_len)
 
         self.basis_to_embed = nn.Sequential(
             nn.Conv1d(num_basis, dim, 1),
@@ -99,24 +207,83 @@ class HSTasNet(Module):
             Rearrange('... (t basis) -> ... basis t', t = num_sources)
         )
 
-        self.conv_decode = nn.ConvTranspose1d(num_basis, audio_channels, segment_len, stride = overlap_len, padding = overlap_len)
+        self.conv_decode = nn.ConvTranspose1d(num_basis, audio_channels, segment_len, stride = overlap_len)
 
         # they do a single layer of lstm in their "small" variant
 
         self.small = small
         lstm_num_layers = 1 if small else 2
 
-        # lstms
+        # rnn
 
-        self.pre_spec_branch = LSTM(dim, dim, lstm_num_layers)
-        self.post_spec_branch = LSTM(dim, dim, lstm_num_layers)
+        rnn_klass = LSTM if not use_gru else GRU
+
+        self.pre_spec_branch = rnn_klass(dim, dim, lstm_num_layers)
+        self.post_spec_branch = rnn_klass(dim, dim, lstm_num_layers)
 
         dim_fusion = dim * (2 if not small else 1)
 
-        self.fusion_branch = LSTM(dim_fusion, dim_fusion, lstm_num_layers)
+        self.fusion_branch = rnn_klass(dim_fusion, dim_fusion, lstm_num_layers)
 
-        self.pre_waveform_branch = LSTM(dim, dim, lstm_num_layers)
-        self.post_waveform_branch = LSTM(dim, dim, lstm_num_layers)
+        self.pre_waveform_branch = rnn_klass(dim, dim, lstm_num_layers)
+        self.post_waveform_branch = rnn_klass(dim, dim, lstm_num_layers)
+
+        # torch compile forward
+
+        if torch_compile:
+            self.forward = torch.compile(self.forward)
+
+    def init_stream_fn(self, device = None):
+        self.eval()
+
+        past_audio = torch.zeros((self.audio_channels, self.overlap_len), device = device)
+        hiddens = None
+
+        @torch.inference_mode()
+        def fn(audio_chunk: ndarray | Tensor):
+            assert audio_chunk.shape[-1] == self.overlap_len
+
+            nonlocal hiddens
+            nonlocal past_audio
+
+            is_numpy_input = isinstance(audio_chunk, ndarray)
+
+            if is_numpy_input:
+                audio_chunk = torch.from_numpy(audio_chunk)
+
+            squeezed_audio_channel = audio_chunk.ndim == 1
+
+            if squeezed_audio_channel:
+                audio_chunk = rearrange(audio_chunk, '... -> 1 ...')
+
+            if exists(device):
+                audio_chunk = audio_chunk.to(device)
+
+            # add past audio chunk
+
+            full_chunk = cat((past_audio, audio_chunk), dim = -1)
+
+            full_chunk = rearrange(full_chunk, '... -> 1 ...')
+
+            # forward chunk with past overlap through model
+
+            transformed, hiddens = self.forward(full_chunk, hiddens = hiddens)
+
+            transformed = rearrange(transformed, '1 ... -> ...')
+
+            if squeezed_audio_channel:
+                transformer = rearrange(transformed, 't 1 ... -> t ...')
+
+            if is_numpy_input:
+                transformed = transformed.cpu().numpy()
+
+            # save next overlap chunk for next timestep
+
+            past_audio = audio_chunk
+
+            return transformed
+
+        return fn
 
     @property
     def num_parameters(self):
@@ -128,7 +295,11 @@ class HSTasNet(Module):
         hiddens = None,
         targets = None
     ):
-        batch, device = audio.shape[0], audio.device
+        batch, audio_len, device = audio.shape[0], audio.shape[-1], audio.device
+
+        assert divisible_by(audio_len, self.segment_len)
+
+        maybe_residual = residual if not self.small else identity
 
         if exists(targets):
             assert targets.shape == (batch, self.num_sources, *audio.shape[1:])
@@ -142,13 +313,17 @@ class HSTasNet(Module):
 
         assert not (self.stereo and audio.shape[1] != 2), 'audio channels must be 2 if training stereo'
 
+        # pad the audio manually on the left side for causal, and set stft center False
+
+        audio = F.pad(audio, (self.causal_pad, 0), value = 0.)
+
         # handle spec encoding
 
         spec_audio_input = rearrange(audio, 'b s ... -> (b s) ...')
 
         stft_window = hann_window(self.segment_len, device = device)
 
-        complex_spec = stft(spec_audio_input, window = stft_window, **self.stft_kwargs, return_complex = True)
+        complex_spec = self.stft(spec_audio_input)
 
         real_spec = view_as_real(complex_spec)
 
@@ -182,9 +357,9 @@ class HSTasNet(Module):
 
         spec_residual, waveform_residual = spec, waveform
 
-        spec, next_pre_spec_hidden = residual(self.pre_spec_branch)(spec, pre_spec_hidden)
+        spec, next_pre_spec_hidden = maybe_residual(self.pre_spec_branch)(spec, pre_spec_hidden)
 
-        waveform, next_pre_waveform_hidden = residual(self.pre_waveform_branch)(waveform, pre_waveform_hidden)
+        waveform, next_pre_waveform_hidden = maybe_residual(self.pre_waveform_branch)(waveform, pre_waveform_hidden)
 
         # if small, they just sum the two branches
 
@@ -195,7 +370,7 @@ class HSTasNet(Module):
 
         # fusing
 
-        fused, next_fusion_hidden = residual(self.fusion_branch)(fusion_input, fusion_hidden)
+        fused, next_fusion_hidden = maybe_residual(self.fusion_branch)(fusion_input, fusion_hidden)
 
         # split if not small, handle small next week
 
@@ -212,9 +387,9 @@ class HSTasNet(Module):
 
         # layer for both branches
 
-        spec, next_post_spec_hidden = residual(self.post_spec_branch)(spec, post_spec_hidden)
+        spec, next_post_spec_hidden = maybe_residual(self.post_spec_branch)(spec, post_spec_hidden)
 
-        waveform, next_post_waveform_hidden = residual(self.post_waveform_branch)(waveform, post_waveform_hidden)
+        waveform, next_post_waveform_hidden = maybe_residual(self.post_waveform_branch)(waveform, post_waveform_hidden)
 
         # spec mask
 
@@ -222,9 +397,9 @@ class HSTasNet(Module):
 
         real_spec_per_source = multiply('b ..., b ... t -> (b t) ...', real_spec, spec_mask)
 
-        complex_spec_per_source = view_as_complex(real_spec_per_source)
+        complex_spec_per_source = view_as_complex(real_spec_per_source.contiguous())
 
-        recon_audio_from_spec = istft(complex_spec_per_source, window = stft_window, **self.stft_kwargs, return_complex = False)
+        recon_audio_from_spec = self.stft.inverse(complex_spec_per_source)
 
         recon_audio_from_spec = rearrange(recon_audio_from_spec, '(b s t) ... -> b t s ...', b = batch, s = self.audio_channels)
 
@@ -247,6 +422,10 @@ class HSTasNet(Module):
         if audio_is_squeezed:
             recon_audio = rearrange(recon_audio, 'b s 1 n -> b s n')
 
+        # excise out the causal padding
+
+        recon_audio = recon_audio[..., self.causal_pad:]
+
         if exists(targets):
             recon_loss = F.l1_loss(recon_audio, targets) # they claim a simple l1 loss is better than all the complicated stuff of past
             return recon_loss
